/* 
 * File:   workerclass.cpp
 * Author: Kristine Ivarsøy and Ida Elise Solhaug
 * 
 * Created on April 3, 2014, 8:27 PM
 */

#include "workerclass.hpp"

#include <cmath>
#include <sstream>


using namespace std;

workerclass::workerclass() {
    
}

workerclass::workerclass(const workerclass& orig) {

}

workerclass::~workerclass() {
}

void workerclass::initialize(char* argv, MPI::Intracomm comm) {
    
    communicator=comm;
    
    rank = communicator.Get_rank();
    
    //allocate memory using the dedicated function
     XPRScreateprob(&prob);
    //read the problem from the file path provided since main
    XPRSreadprob(prob,argv,"");
    
    // create logfile
    string logfile;
    stringstream ss;
    ss << rank << "logfile.txt";
    logfile = ss.str();
    XPRSsetlogfile(prob, logfile.c_str());
    
    //Get number of variables in the problem
    XPRSgetintattrib(prob, XPRS_COLS, &nCols);
   
    char *types = new char[nCols];
    int *temp_bin_indices = new int[nCols];
    
    //Store type of each variable in the problem
    XPRSgetcoltype(prob, types, 0, nCols - 1);
    
    num_bins=0;
    // Count number of binary variables in the problem
    for (int i = 0; i < nCols; i++) {
        if (types[i] == 'B') {
            temp_bin_indices[num_bins] = i;
            num_bins++;
            //cout<<" index of binary variable number "<<num_bins<<" is "<<i<<endl;
        }
    }
    
    //copy the indices into the proper field and destroy the temporal array
    indexBinVar = new int[num_bins];
    for(int i=0;i<num_bins;i++){
        indexBinVar[i] = temp_bin_indices[i];
    }
    
    delete[] temp_bin_indices;
    delete[] types;
}

int workerclass::optimize(){
    //receive from the master a notification of what to do: 1 to solve, 0 to terminate
    receive_notification();

    //if I am told to solve
    while (instruction != 0){
    
        //receive from the master a set of fixed variables
        binaries = new int[num_bins];

        receive_binaries();
        
        //change the upper bounds according to the fixing vector
        double *bounds = new double[2*num_bins];
        char *type_of_bound = new char[2*num_bins];
        int *var_index = new int[2*num_bins];
        
        int nn = 0;
        
        //fill the array to change the upper bounds:
        for(int i=0;i<num_bins;i++){
            if (binaries[i] == -1) {
                type_of_bound[nn]='U';
                bounds[nn] = 1;
                var_index[nn] = indexBinVar[i];
                nn++;
                type_of_bound[nn]='L';
                bounds[nn] = 0;
                var_index[nn] = indexBinVar[i];
                nn++;
            }
            else if (binaries[i] == 1) {
                type_of_bound[nn]='B';
                bounds[nn] = 1;
                var_index[nn] = indexBinVar[i];
                nn++;

            }
            
            else if (binaries[i] == 0) {
                type_of_bound[nn]='B';
                bounds[nn] = 0;
                var_index[nn] = indexBinVar[i];
                nn++;
            }
            
        }
        
        XPRSchgbounds(prob, nn, var_index, type_of_bound, bounds);

        delete[] type_of_bound;
        delete[] bounds;
        delete[] var_index;
 
       
        //Solve the LP relaxed problem
        XPRSlpoptimize(prob, "");

        // Get status of the solution
        XPRSgetintattrib(prob,XPRS_LPSTATUS,&solution_status);
       
        // Check if the solution is infeasible
        if (solution_status == 2) {
            solution_status = 0; 
            cout<<rank<<": Problem is infeasible! "<<endl;
        }
        
        // Check if the solution is LP optimal
        else if (solution_status == 1) {
            int int_feas = 1;
            
            // Read the variable values in the solution
            x = new double[nCols];
            XPRSgetlpsol(prob, x, NULL, NULL, NULL);
            
            //Set initial branch_variable to the first binary variable
            branch_variable = 0;
            

            // Check if some binary values are not 0 or 1
            for (int i = 0; i < num_bins; i++) {
                if (x[indexBinVar[i]] > 0.0001 and x[indexBinVar[i]] < 0.9999) {

                    // Find the variable with solution value closest to 0.5 and set as branch variable
                    if (abs(x[indexBinVar[i]]-0.5) < abs(x[indexBinVar[branch_variable]]-0.5)) {
                        branch_variable=i;
                    }
                   //If any of the variables are strictly between upper and lower bounds we no not have an integer feasible solution 
                   int_feas = 0; 
                }
            }
            
            // Check if the solution is integer feasible
            if (int_feas == 1) {
                solution_status = 2;// solution_status=2 indicates an integer feasible solution
                cout<<"Integer solution found!"<<endl;
            }
            
            // if not it is a lower bound
            else {
             solution_status = 1; // solution_status=1 indicates an LP optimal solution that is not integer feasible
            } 
            
            // Save the objective value
            XPRSgetdblattrib(prob, XPRS_LPOBJVAL, &solution_value);
            
            delete[] x;

        }
        
        // if LPSTATUS is different from 1 or 2, something is wrong...Needs some kind of error handling at receiver
        else {
             cout<<rank<<": Something weird happened in xpress, returned lpstatus "<<solution_status<<endl;
             for(int i=0;i<num_bins;i++){
               cout<<" Binary i: "<<binaries[i]<<endl;

            }
             solution_status = 3;
             return_solution();
             break;
        }    
        
        delete[] binaries;

        
        //Return the solution to the master
        return_solution();
        
       
        
        // Receive new instruction
        receive_notification();

    }
    //else I am told to terminate
    finalize();
    return 0;
    
}

void workerclass::receive_notification(){
    communicator.Recv(&instruction, 1, MPI::INT, 0, 0, status);
}

void workerclass:: receive_binaries() {
    communicator.Recv(binaries, num_bins, MPI::INT, 0, 1, status);
}

void workerclass::return_solution(){
    request=communicator.Issend(&solution_status, 1, MPI::INT, 0, 2);
    request.Wait(status);
    // If solution is LP optimal, return value and branch variable
    if (solution_status==1){
        request=communicator.Issend(&solution_value, 1, MPI::DOUBLE, 0, 3);

        request.Wait(status);

        request=communicator.Issend(&branch_variable, 1, MPI::INT, 0, 4);
        request.Wait(status);

    }
    // if solution is integer feasible, return value
    else if (solution_status==2) {
        request=communicator.Issend(&solution_value, 1, MPI::DOUBLE, 0, 3);
        request.Wait(status);
    }
}

void workerclass::finalize(){
    delete[] indexBinVar;
    XPRSdestroyprob(prob);
}