/* 
 * File:   bbclass.cpp
 * Author: Kristine Ivarsøy and Ida Elise Solhaug
 * 
 * Created on April 3, 2014, 8:26 PM
 */

#include "bbclass.hpp"
#include <cstdio>
#include <vector>
#include <math.h>       /* log2 */
#include <float.h>

using namespace std;

bbclass::bbclass() {
}

bbclass::bbclass(const bbclass& orig) {
}

bbclass::~bbclass() {
}


void bbclass::initialize(char *argv, MPI::Intracomm comm){
    
    communicator = comm;
    
    myRank = communicator.Get_rank();
    
    numProc=communicator.Get_size();

    cout<<"NumProc is "<<numProc<<endl;
    
    //declare an xpress problem object
    XPRSprob prob;
    //allocate memory using the dedicated function
    XPRScreateprob(&prob);
    //read the problem from the file path provided since main
    XPRSreadprob(prob, argv, "");
    //read the number of binary variables in the problem
    int nCols = 0;
    XPRSgetintattrib(prob, XPRS_COLS, &nCols);
    
    char *types = new char[nCols];
    num_bins=0;
    XPRSgetcoltype(prob, types, 0, nCols-1);
    for(int i=0;i<nCols;i++) if (types[i]=='B') num_bins++;

    incumbent_worker = 0;
    incumbent_value = DBL_MAX;
    
    delete [] types;
    XPRSdestroyprob(prob);
}

void intToBin(int n, vector<int> &bin){
    int d=n;
    
    if (n > 1) {
        d = n % 2;
        intToBin(n / 2, bin);
    }
    
    bin.push_back(d);
}

int bbclass::optimize() {

    // Count how many subtrees have been created
    int node_counter = 0;
    
    // Number of variables to fix
    int fixed_vars = log2 (numProc-1);
    
    // Array to send to worker
    int *fix_relax = new int[num_bins]; 

    // Arrays to hold solution status and values
    solution_status = new int[numProc];
    solution_value = new double[numProc];

    cout<<"Number of fixed variables is "<<fixed_vars<<endl;
    
    // Vector to fill in binary representation of a number i up to numProc
    vector<int> bin;
    
    for(int i=0; i<numProc-1;i++) {

       // Fill in vector with binary representation of number i
       intToBin(i, bin);

      // Fill vector with 0s before the binary representation
       while(bin.size()<fixed_vars){
           bin.insert(bin.begin(),0);
       }

       // Fill vector with -1 after the binary representation
       while(bin.size()<num_bins){
           bin.push_back(-1);
       }

       node_counter++;
       
       // Fix variables up to fixed_vars
       int fixed = 0;
       
       fix_relax[0]= -1;
       
       for(int i=0; i<num_bins-1; i++) {
           // Fix every 50th variable
           if (i % 40 == 0 and fixed<fixed_vars) {
            fix_relax[i+1]=bin[fixed];
                   cout<<"Binary variable number "<<i+1<<" is fixed to "<<fix_relax[i+1]<<endl;
            fixed++;
            }
           else {
            fix_relax[i+1]= -1;
           }
       }
      
       // Send notification
        send_notification(1, node_counter);
                        cout<<"Sent notification to "<<node_counter<<endl;

        send_binaries(fix_relax, node_counter);
                                cout<<"Sent binaries to "<<node_counter<<endl;

        delete[] fix_relax;
        bin.clear();
    }

    // Count received solutions
    int count_received = 0;
    
        // While not all workers have returned
        while (count_received < numProc-1) {
            // Wait until something is received
            while (communicator.Iprobe(MPI::ANY_SOURCE, MPI::ANY_TAG, status) == 0);
            // Save identity of origin worker
            int origin_worker = status.Get_source();

            // Receive the solution status and value if any
            receive_solution(origin_worker);
            
            cout<<"Received solution from : "<<origin_worker<<endl;

            
            if (solution_status[origin_worker]==1) {
                if(solution_value[origin_worker]<incumbent_value) {
                    incumbent_value = solution_value[origin_worker];
                    incumbent_worker = origin_worker;
                    cout<<"New best solution found!"<<endl;
                }
            }
               
            count_received++;

            cout<<"Have received "<<count_received<<" solutions now "<<endl;
            cout<<"Current best solution is : "<<incumbent_value<<" received from "<<incumbent_worker<<endl;
        }
    
        cout<<"Everyone is done optimizing!"<<endl;
        cout<<"Best solution is : "<<incumbent_value<<" received from "<<incumbent_worker<<endl;

        delete [] solution_value;
        delete [] solution_status;
    
        return 0;
}

// Function to send notification to worker, 0 means terminate, 1 means solve
void bbclass::send_notification(int notification, int proc) {
    request = communicator.Issend(&notification, 1, MPI::INT, proc, 0);
    request.Wait(status);
}

// Function to send the status of binary variables to a worker
void bbclass::send_binaries(int* binaries, int proc) {
    request = communicator.Issend(binaries, num_bins , MPI::INT, proc, 1);
    request.Wait(status);
}

// Function to receive solution_status from a worker, 0 means infeasible solution, 1 means LP feasible solution and 2 means integer feasible solution
void bbclass::receive_solution(int origin_worker) {
    communicator.Recv(&solution_status[origin_worker], 1, MPI::INT, origin_worker, 2, status);
    // If integer solution was found, receive value as well
    if (solution_status[origin_worker]==1) {
        communicator.Recv(&solution_value[origin_worker], 1, MPI::DOUBLE, origin_worker, 3, status);
    }
}
